/*
 *	xled: multi-purpose LED display for the X11 desktop
 *      
 *      this software is free to use and comes with absolutley no
 *      warranty.  
 *      Copyright (c) 1996,1997 Jan Schoenepauck & Joachim Gassen
 *
 *      For comments, bug reports, and requests please contact 
 *
 *       schoenep@uni-wuppertal.de
 *       joachim@fb4-1112.uni-muenster.de
 *      
 *      aka Fast Forward Productions
 *
 *      For further information: http://fb4-1112.uni-muenster.de/ffwd
 *       
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <string.h>
#include <dirent.h>
#include <linux/dirent.h>
#include <linux/unistd.h>
#include <linux/limits.h>
#include <sys/stat.h>
#include <unistd.h>
#include <getopt.h>
#include <setjmp.h>
#include <math.h>
#include <sys/wait.h>
#include <sys/types.h>

#include "global.h"
#include "xinit.h"
#include "colors.h"
#include "windows.h"
#include "config.h"
#include "kbd_leds.h"
#include "mail.h"

#define MIN(a,b) (((a)<(b))?(a):(b))

char *x_display = NULL;
char *x_geom = NULL;
int geom_flags;
int have_position = 0;
int grav;
int winx = 0, winy = 0, winw = 0, winh = 0;
char *fg_color, *bg_color, *lo_color, *zf_color, *zl_color;
int check_period=0;
char *incoming_dir, *exec_command, *mail_spool_file, *back_end_command;
char *title;
int alarm_counter=0, call_count, old_call_count = -1;
int mail_mode=0;
int front_end_mode=0;
int vbox_mode=0;
int bell_mode=0;
int digits=0;
int twocolor_mode=0;
static jmp_buf prog;
int show_all_messages = 0;
static char *default_font_name = "-*-helvetica-medium-r-*-*-12-*";
FILE *mail_file; 
Win mainwin;

Pixmap icon, icon_mask;
#include "xled.xbm"
#include "xled_mask.xbm"

/* This exit function must be called at the end of the program; it also */
/* serves as a signal handler for SIGTERM                               */
void close_all (int s) {
  if ((use_kbd_leds) && (!kbd_leds_cleared)) clear_kbd_leds();
  disconnect_server();
  if (s<0) {
    printf("%s: abnormal program termination.\n",myname);
    exit(-1);
  }
  if (s==0) printf("%s: terminated by user.\n",myname);
  else printf("%s: exiting on signal %d.\n", myname, s);  
  exit(0);
}                  


/* this function is based on the fact that the x flag of the messages 
   is being toggled when the messages have been played by vbox. */ 

int check_vbox_messages (void) {
  struct dirent *dir_info;
  DIR *dir;
  int count=0;
  struct stat stat_buf; 
  char file_name[PATH_MAX+1];
  dir = opendir (incoming_dir);
  if (dir == NULL) {
    fprintf (stderr, "%s: Error opening vbox directory: %s\n", 
	     myname, strerror(errno));
    return (-1);
  }
  dir_info = readdir (dir);
  while (dir_info != NULL) {
    sprintf(file_name,"%s/%s",incoming_dir,dir_info->d_name);
    if ((dir_info->d_name[0]!='.')&&(stat (file_name,&stat_buf) >= 0)) { 
      if (show_all_messages) {
	if (stat_buf.st_mode&S_IREAD) count ++;
      }	else 
#ifdef _NEW_VBOX_VERSION
	if ((stat_buf.st_mode&S_IREAD)&&stat_buf.st_mtime) count ++;
#else
	if ((stat_buf.st_mode&S_IREAD)&&!(stat_buf.st_mode&S_IXUSR)) count ++;
#endif    
    }
    dir_info = readdir (dir);
  }
  closedir(dir);
  return (count);
}

int get_count (void) {
  sigset_t  sig_set_new, sig_set_old;
  int ret_val;
  if (mail_mode) return (count_mail(mail_spool_file,show_all_messages));
  else if (front_end_mode) {
    if ((ret_val=fork())==0) {
      signal(SIGALRM,SIG_DFL);
      execl("/bin/sh", "sh", "-c", back_end_command, NULL);
      _exit(127);
    } 
    else {
      sigemptyset(&sig_set_new);
      sigaddset(&sig_set_new,SIGALRM);
      sigprocmask(SIG_BLOCK,&sig_set_new,&sig_set_old);
      wait(&ret_val);  
      sigprocmask(SIG_SETMASK,&sig_set_old,NULL);    
    }
    return(ret_val>>8);
  }	
  return(check_vbox_messages ());
}


int led_lit (int pos, int val) {
  static int lookup[10]= {125,80,55,87,90,79,111,81,127,95};
  return ((1<<pos)&lookup[val]);
}
  

void draw_single_led (int start_x, int start_y, int x, int y, int w, int val,
		      PIXEL hi, PIXEL lo) {
  int i;
  for (i=0;i<w;i++) {
    if (led_lit(0,val)) XSetForeground(display,mainwin.gc,hi);
    else XSetForeground(display,mainwin.gc,lo);
    XDrawLine (display, mainwin.window, mainwin.gc, 
	       start_x+1+i, start_y+i, start_x+x-2-i, start_y+i);
    if (led_lit(1,val)) XSetForeground(display,mainwin.gc,hi);
    else XSetForeground(display,mainwin.gc,lo);
    if (!(i%2)) XDrawLine (display, mainwin.window, mainwin.gc,
			   start_x+2+(i/2),   start_y+y-(i/2)+1,
			   start_x+x-3-(i/2), start_y+y-(i/2)+1);
    else     XDrawLine (display, mainwin.window, mainwin.gc,
			start_x+2+(i/2)+1,       start_y+y+(i/2)+2,
			start_x+x-3-((i/2)+1), start_y+y+(i/2)+2);
    if (led_lit(2,val)) XSetForeground(display,mainwin.gc,hi);
    else XSetForeground(display,mainwin.gc,lo);
    XDrawLine (display, mainwin.window, mainwin.gc,
	       start_x+1+i,   start_y+2*y-i+2, 
	       start_x+x-2-i, start_y+2*y-i+2);
    if (led_lit(3,val)) XSetForeground(display,mainwin.gc,hi);
    else XSetForeground(display,mainwin.gc,lo);
    XDrawLine (display, mainwin.window, mainwin.gc, 
	       start_x+i, start_y+2+i, start_x+i, start_y+y-i);
    if (led_lit(4,val)) XSetForeground(display,mainwin.gc,hi);
    else XSetForeground(display,mainwin.gc,lo);
    XDrawLine (display, mainwin.window, mainwin.gc,
	       start_x+x-i-1, start_y+2+i, start_x+x-i-1, start_y+y-i);
    if (led_lit(5,val)) XSetForeground(display,mainwin.gc,hi);
    else XSetForeground(display,mainwin.gc,lo);
    XDrawLine (display, mainwin.window, mainwin.gc,
	       start_x+i, start_y+2+i+y, start_x+i, start_y+2*y-i);
    if (led_lit(6,val)) XSetForeground(display,mainwin.gc,hi);
    else XSetForeground(display,mainwin.gc,lo);
    XDrawLine (display, mainwin.window, mainwin.gc,
	       start_x+x-i-1, start_y+2+i+y, start_x+x-1-i, start_y+2*y-i);
  }
}


void draw_led (int val, int dig, int win_size_x, int win_size_y) {
  int x,y,w,i;
  int xt,yt,zero;
  XSetForeground(display,mainwin.gc,window_bg);
  XFillRectangle(display,mainwin.window,mainwin.gc,0,0,win_size_x,win_size_y);
  x = win_size_x / ((float)(dig + (dig+1)*0.25));
  y = win_size_y / 2.5;
  w = MIN(x,y) /6;
  if (w<3) w=3;
  if (!(w%2)) w++;
  xt = x/4;
  yt = y/4;
  if (val>(pow(10,digits)-1)) val = pow(10,dig)-1;
  if (val<0)  val = 0;
  zero=twocolor_mode?(val?0:1):0;
  for (i=0;i<digits-1;i++) {
    if (!zero) draw_single_led (xt,yt,x,y,w,val/((int)pow(10,dig-i-1)),
			      led_on,led_off);
    else  draw_single_led (xt,yt,x,y,w,val/((int)pow(10,dig-i-1)),
			      z_led_on,z_led_off);
    xt = (0.25+1.25*(i+1)) * x;
    val=val%((int)pow(10,dig-i-1));
  }
  if (!zero) draw_single_led (xt,yt,x,y,w,val,led_on,led_off);
  else  draw_single_led (xt,yt,x,y,w,val,z_led_on,z_led_off);
}

void update_mainwindow (void) {
  old_call_count=call_count;
  draw_led(call_count,digits,mainwin.w,mainwin.h);
}


void resize_mainwindow (XEvent *event) {
  int new_w, new_h;
  new_w = event->xconfigure.width;
  new_h = event->xconfigure.height;
  mainwin.w = new_w;
  mainwin.h = new_h;
}

/* This is called in the main loop: checks for queued events from the X */
/* server, and calls the appropriate handlers if events were received.  */
void check_events (void) {
  XEvent event;
  int ret_val;
  if (setjmp(prog) != 0) 
    if (call_count != old_call_count) {
      if ((!old_call_count)&&(bell_mode)) XBell(display,100);
      update_mainwindow();
    }
  XNextEvent(display, &event);
  switch (event.type) {
  case Expose :
    discard_exposures(mainwin.window);
    update_mainwindow();
    break;
  case ButtonPress : 
    if ((ret_val=fork())==0) {
      signal(SIGALRM,SIG_DFL);
      execl("/bin/sh", "sh", "-c", exec_command, NULL);
      _exit(127);
    } 
    else {
      setjmp(prog);
      wait(&ret_val);  
    }
    break;
  case ConfigureNotify :
    resize_mainwindow(&event);
    break;
  case DestroyNotify :
    close_all(0);
    break;
  default :
    break;
  }
}


static void usage () {
  static char *help_message[] = {
    "where options include:",
    "    -vbox                check for voice messages from vboxgetty",
    "    -mail                check for new mail messages",
    "    -frontend            use xled as a frontend to display an exit code",
    "    -display host:dpy    X server to contact",
    "    -geometry geom       size of display",
    "    -bg color            background color",
    "    -fg color            foreground color",
    "    -lo color            color of non-lit LEDs",    
    "    -twocolor            use a different color if value displayed is zero",
    "    -zf color            color of non-lit LEDs for zero in twocolor mode",
    "    -zl color            color of lit LEDs for zero in twocolor mode",
    "    -bell                ring the bell if the LED increases from zero",
    "    -digits count        count of digits to show",
    "    -kbdled mask         binary mask of keyboard leds to blink",
    "    -all                 show all messages, not only the new ones",
    "    -update seconds      how often to check for new incoming messages",
    "    -vpath path          where to look for new vboxgetty messages",
    "    -mfile file          user's mail spool-file",
    "    -exec command        commandline to execute on button click",
    "    -backend command     commando which exit code is to be displayed",
    "    -title string        title string of the xled window", 
    "    -help                display this screen and exit",
    NULL
  };
  char **cpp;
  fprintf(stderr, "xled v%s\n\n",VERSION);
  fprintf(stderr, "usage:  %s [-vbox | -mail | -frontend] [-options ...]\n", 
	  myname);
  for (cpp = help_message; *cpp; cpp++) {
    fprintf (stderr, "%s\n", *cpp);
  }
  fprintf (stderr, "\n");
  exit (1);
}

void parse_command_line (int argc, char **argv) {
  struct command_options {
    char *command;
    int arg_count;
  } options[] = {
    {"-display",1},
    {"-geometry",1},
    {"-bg",1},
    {"-fg",1},
    {"-lo",1},
    {"-update",1 },
    {"-vpath",1 },
    {"-exec",1 },
    {"-help",0},
    {"-all",0},
    {"-kbdled",1 },
    {"-mail",0},
    {"-mfile",1},
    {"-frontend",0},
    {"-backend",1},
    {"-digits",1},
    {"-vbox",0},
    {"-title",1},
    {"-twocolor",0},
    {"-zf",1},
    {"-zl",1},
    {"-bell",0},
    {NULL,0}
  };
  int i,j,incorrect_arg;
  /* First get my own program name and put it in <myname> */
  myname = strrchr(argv[0], '/');
  if (myname == NULL) myname = argv[0];
  else myname++;
  for (i=1;i<argc;i++) {
    j = 0;
    incorrect_arg=1;
    while (options[j].command) {
      if (strcmp(argv[i],options[j].command)==0) {
	if (options[j].arg_count) {
	  if (argc<=i+1) usage();
	}
	switch (j) {
	case 0 : x_display = argv[++i];
	  incorrect_arg=0;
	  break;
	case 1 : x_geom = argv[++i];
	  incorrect_arg=0;	
	  break;
	case 2 : bg_color = argv[++i];
	  incorrect_arg=0;
	  break;
	case 3 : fg_color = argv[++i];
	  incorrect_arg=0;
	  break;
	case 4 : lo_color = argv[++i];
	  incorrect_arg=0;
	  break;
	case 5 : if ((check_period = atoi (argv [++i])) <= 0) usage ();
	  incorrect_arg=0;
	  break;
	case 6 : incoming_dir = argv [++i];
	  incorrect_arg=0;
	  break;
	case 7 : exec_command = argv [++i];
	  incorrect_arg=0;
	  break;  
	case 8 : usage ();
	  break;
	case 9 : show_all_messages = 1;
	  incorrect_arg=0;
	  break;
	case 10 : if ((kbd_led_mask = atoi(argv [++i]))<=0) usage();
	  use_kbd_leds = 1;
	  incorrect_arg=0;
	  break;
	case 11 : mail_mode = 1;
	  incorrect_arg=0;
	  break;
	case 12 : mail_spool_file = argv [++i];
	  incorrect_arg=0;
	  break;
	case 13 : front_end_mode = 1;
	  incorrect_arg=0;
	  break;
	case 14 : back_end_command = argv [++i];
	  incorrect_arg=0;
	  break;
	case 15 : if ((digits = atoi(argv [++i]))<=0) usage();
	  incorrect_arg=0;
	  break;
	case 16 : vbox_mode = 1;
	  incorrect_arg=0;
	  break;
	case 17 : title = argv [++i];
	  incorrect_arg=0;
	  break;
	case 18 : twocolor_mode = 1;
	  incorrect_arg=0;
	  break;
	case 19 : zf_color = argv[++i];
	  incorrect_arg=0;
	  break;
	case 20 : zl_color = argv[++i];
	  incorrect_arg=0;
	  break;
	case 21 : bell_mode = 1;
	  incorrect_arg=0;
	  break;	  
	default : break;
	}
      }
      j++;
    }
    if (incorrect_arg) usage();
  }
  if ((DEFAULT_MODE==1)&&(mail_mode&&front_end_mode)) usage();
  if ((DEFAULT_MODE==1)&&((!mail_mode)&&(!front_end_mode))) vbox_mode=1;
  if ((DEFAULT_MODE==2)&&(vbox_mode&&front_end_mode)) usage();
  if ((DEFAULT_MODE==2)&&((!vbox_mode)&&(!front_end_mode))) mail_mode=1;
  if ((DEFAULT_MODE==3)&&(vbox_mode&&mail_mode)) usage();
  if ((DEFAULT_MODE==3)&&((!vbox_mode)&&(!mail_mode))) front_end_mode=1;
  if (mail_spool_file==NULL) {
    mail_spool_file = (char *) malloc (strlen (MAIL_SPOOL_FILE)+1);
    strcpy (mail_spool_file,MAIL_SPOOL_FILE);
  }
  if (incoming_dir==NULL) {
    incoming_dir = (char *) malloc (strlen (INCOMING_DIR)+1);
    strcpy (incoming_dir,INCOMING_DIR);
  }
  if (check_period==0) check_period = CHECK_PERIOD;
  if (fg_color==NULL) {
    fg_color = (char *) malloc (strlen (FG_COLOR) +1);
    strcpy (fg_color,FG_COLOR);
  }
  if (bg_color==NULL) {
    bg_color = (char *) malloc (strlen (BG_COLOR) +1);
    strcpy (bg_color,BG_COLOR);
  }
  if (lo_color==NULL) {
    lo_color = (char *) malloc (strlen (LO_COLOR) +1);
    strcpy (lo_color,LO_COLOR);
  }
  if (exec_command==NULL) {
    exec_command = (char *) malloc (strlen (EXEC_COMMAND)+1);
    strcpy(exec_command, EXEC_COMMAND);
  }
  if (back_end_command==NULL) {
    back_end_command = (char *) malloc (strlen (BACK_END_COMMAND)+1);
    strcpy (back_end_command, BACK_END_COMMAND);
  }
  if (title==NULL) {
    title = (char *) malloc (strlen (TITLE)+1);
    strcpy (title, TITLE);
  }
  if (!digits) digits=DIGITS;
  if (twocolor_mode) {
    if (zf_color==NULL) {
      zf_color = (char *) malloc (strlen (ZF_COLOR) +1);
      strcpy (zf_color,ZF_COLOR);
    }
    if (zl_color==NULL) {
      zl_color = (char *) malloc (strlen (ZL_COLOR) +1);
      strcpy (zl_color,ZL_COLOR);
    }
  }
}


/* This function is called when a SIGALRM signal is received. Checks for */
/* new messages in the incoming messages path and updates the counter    */
/* display if the number changed. The SIGALRM signal is raised in        */
/* regular intervals (check_period, in seconds) as set by the alarm()    */
/* function.                                                             */
void alarm_handler (int sig) {
  if (alarm_counter++==check_period) {
    call_count = get_count ();
    alarm_counter=0;
  }
  if (call_count < 0) close_all (-1);
  if (use_kbd_leds) {
    if (call_count) toggle_kbd_leds(); 
    else if (!kbd_leds_cleared) clear_kbd_leds();
  }
  signal(SIGALRM,alarm_handler);
  alarm(1);
  longjmp (prog ,1);
}




int main(int argc, char **argv) {
  parse_command_line(argc,argv);
/* Basic initialization - try connecting the X server */
  connect_server(x_display);
/* set up the signal handlers */
  signal(SIGTERM,close_all);
  signal(SIGINT,close_all);
  signal(SIGALRM,alarm_handler);

  init_cmap();
#ifdef NO_COLOR_ERRORS
  alloc_named_color(cmap,bg_color,&window_bg);
  alloc_named_color(cmap,fg_color,&led_on);
  alloc_named_color(cmap,lo_color,&led_off);
  if (twocolor_mode) {
    alloc_named_color(cmap,zf_color,&z_led_on);
    alloc_named_color(cmap,zl_color,&z_led_off);
  }
#else
  if (!alloc_named_color(cmap,bg_color,&window_bg)) {
    fprintf(stderr,"%s: couldn't allocate color %s\n",myname,bg_color);
    exit(-1);
  }
  if (!alloc_named_color(cmap,fg_color,&led_on)) {
    fprintf(stderr,"%s: couldn't allocate color %s\n",myname,fg_color);
    exit(-1);
  }
  if (!alloc_named_color(cmap,lo_color,&led_off)) {
    fprintf(stderr,"%s: couldn't allocate color %s\n",myname,lo_color);
    exit(-1);
  }
  if (twocolor_mode) {
    if (!alloc_named_color(cmap,zf_color,&z_led_on)) {
      fprintf(stderr,"%s: couldn't allocate color %s\n",myname,zf_color);
      exit(-1);
    }
    if (!alloc_named_color(cmap,zl_color,&z_led_off)) {
      fprintf(stderr,"%s: couldn't allocate color %s\n",myname,zl_color);
      exit(-1);
    }
  }
#endif
  if (use_kbd_leds) {
    init_kbd_leds();
  }

  if (x_geom) geom_flags = XParseGeometry(x_geom, &winx, &winy, &winw, &winh);
  if (winw <= 0) {
    if (winh <= 0) winh = 100;
    winw = (digits * winh) / 2;
  }
  else if (winh <= 0) winh = (winw * 2) / digits;
  grav = NorthWestGravity;
  if ((geom_flags & XValue) && (geom_flags & YValue)) {
    have_position = 1;
    if (geom_flags & XNegative) {
      winx += DisplayWidth(display, screen) - winw;
      if (geom_flags & YNegative) {
        winy += DisplayHeight(display, screen) - winh;
        grav = SouthEastGravity;
      }
      else grav = NorthEastGravity;
    }
    else if (geom_flags & YNegative) {
      winy += DisplayHeight(display, screen) - winh;
      grav = SouthWestGravity;
    }
  }
  if (!load_font(&default_font,default_font_name)) exit(-1);
  icon = XCreateBitmapFromData(display, root, xled_bits, 
                               xled_width, xled_height);
  icon_mask = XCreateBitmapFromData(display, root, xled_mask_bits, 
                                    xled_mask_width, xled_mask_height);
  create_window(&mainwin, root, have_position, winx, winy, winw, winh, grav,
                title, window_bg, led_on, default_font, icon, icon_mask);
  XSelectInput(display, mainwin.window, 
	       ExposureMask|ButtonPressMask|StructureNotifyMask);
  map_window(&mainwin,1);

  call_count = get_count ();
  if (call_count < 0) close_all (-1);
  alarm(check_period);
  for (;;) {
    check_events();
  }
  close_all(0);
  return (0);
}
